package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.MenuProvider
import java.lang.Runnable
import net.minecraft.server.level.ServerLevel
import net.minecraft.client.multiplayer.ClientLevel
import net.minecraft.core.SectionPos
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.StringTag
import net.minecraft.network.chat.Component
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.RandomizableContainerBlockEntity
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.nbt.ifHas
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.oncePre

abstract class MatteryBlockEntity(
	p_155228_: BlockEntityType<*>,
	p_155229_: BlockPos,
	p_155230_: BlockState
) : SynchronizedBlockEntity(p_155228_, p_155229_, p_155230_), MenuProvider, IRedstoneControlProvider {
	var customDisplayName: Component? = null
	override val redstoneControl: IRedstoneControl = RedstoneControl(::redstoneStatusUpdated)

	protected open val defaultDisplayName: Component
		get() = level?.getBlockState(blockPos)?.block?.name ?: TextComponent("null at $blockPos")

	abstract override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu?

	protected fun tickOnce(func: Runnable) {
		level?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceServer(func: Runnable) {
		(level as? ServerLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnceClient(func: Runnable) {
		(level as? ClientLevel)?.oncePre { if (!isRemoved) func.run() }
	}

	protected fun tickOnce(func: (Level) -> Unit) {
		val level = level
		level?.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceServer(func: (ServerLevel) -> Unit) {
		val level = level as? ServerLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	protected fun tickOnceClient(func: (ClientLevel) -> Unit) {
		val level = level as? ClientLevel ?: return
		level.oncePre { if (!isRemoved) func.invoke(level) }
	}

	override fun getDisplayName(): Component {
		return customDisplayName ?: defaultDisplayName
	}

	protected open fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		val customDisplayName = customDisplayName
		if (customDisplayName != null)
			nbt.putString("Name", Component.Serializer.toJson(customDisplayName))

		nbt[REDSTONE_CONTROL_KEY] = redstoneControl.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas("Name", StringTag::class.java) {
			customDisplayName = Component.Serializer.fromJson(it.asString)
		}

		redstoneControl.deserializeNBT(nbt[REDSTONE_CONTROL_KEY] as? CompoundTag)
	}

	// Just to mark chunk unsaved
	open fun setChangedLight() {
		val level = level
		if (level is ServerLevel) {
			level.chunkSource.getChunkNow(
				SectionPos.blockToSectionCoord(blockPos.x),
				SectionPos.blockToSectionCoord(blockPos.z))?.isUnsaved = true
		}
	}

	companion object {
		const val REDSTONE_CONTROL_KEY = "redstoneControl"
		const val INVENTORY_KEY = "container"
		const val MATTER_STORAGE_KEY = "matterStorage"
		const val ENERGY_KEY = "energyStorage"
		const val BATTERY_KEY = "batteryInventory"
		const val LOOT_TABLE_KEY = RandomizableContainerBlockEntity.LOOT_TABLE_TAG
		const val LOOT_TABLE_SEED_KEY = RandomizableContainerBlockEntity.LOOT_TABLE_SEED_TAG
	}
}
