package ru.dbotthepony.mc.otm.client.render.blockentity

import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.client.renderer.MultiBufferSource
import net.minecraft.client.renderer.blockentity.BlockEntityRenderer
import net.minecraft.client.renderer.blockentity.BlockEntityRendererProvider
import net.minecraft.core.Direction
import ru.dbotthepony.mc.otm.block.RotatableMatteryBlock
import ru.dbotthepony.mc.otm.block.entity.tech.BatteryBankBlockEntity
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.matter.MatterCapacitorBankBlockEntity
import ru.dbotthepony.mc.otm.client.render.AbstractMatterySprite
import ru.dbotthepony.mc.otm.client.render.DynamicBufferSource
import ru.dbotthepony.mc.otm.client.screen.widget.MatterGaugePanel
import ru.dbotthepony.mc.otm.client.screen.widget.PowerGaugePanel
import ru.dbotthepony.mc.otm.core.get
import ru.dbotthepony.mc.otm.core.math.rotateY
import kotlin.math.PI

abstract class BankRenderer<T : MatteryBlockEntity>(private val context: BlockEntityRendererProvider.Context) : BlockEntityRenderer<T> {
	protected abstract fun gaugeLevel(entity: T): Float
	protected abstract val texture: AbstractMatterySprite

	override fun render(
		blockEntity: T,
		p_112308_: Float,
		stack: PoseStack,
		p_112310_: MultiBufferSource,
		p_112311_: Int,
		p_112312_: Int
	) {
		val plainLevel = gaugeLevel(blockEntity)

		if (plainLevel <= 0f) {
			return
		}

		stack.pushPose()

		val facing = blockEntity.blockState[RotatableMatteryBlock.FACING]
		val rotateFacing = facing == Direction.NORTH || facing == Direction.SOUTH

		if (rotateFacing) {
			stack.rotateY(PI.toFloat() / 2f)
		}

		stack.scale(0.02f, 0.01f, 0.01f)
		stack.translate(if (rotateFacing) -50.0 else 0.0, 0.0, -0.5)

		// ~50x100 canvas

		val width = 9f
		val heightMax = 39.36f

		val buffer = DynamicBufferSource.WORLD.getBuffer(texture.renderTypeWorld)

		texture.uploadOntoPartialColor(
			stack,
			buffer,
			x = 25f - width / 2f,
			y = 30f,
			width = width,
			height = heightMax * gaugeLevel(blockEntity)
		)

		stack.rotateY(PI.toFloat())
		stack.translate(-50.0, 0.0, -101.0)

		texture.uploadOntoPartialColor(
			stack,
			buffer,
			x = 25f - width / 2f,
			y = 30f,
			width = width,
			height = heightMax * gaugeLevel(blockEntity)
		)

		stack.popPose()
	}
}

class BatteryBankRenderer(context: BlockEntityRendererProvider.Context) : BankRenderer<BatteryBankBlockEntity>(context) {
	override fun gaugeLevel(entity: BatteryBankBlockEntity): Float {
		return entity.gaugeLevel
	}

	override val texture: AbstractMatterySprite
		get() = PowerGaugePanel.GAUGE_FOREGROUND
}

class MatterBatteryBankRenderer(context: BlockEntityRendererProvider.Context) : BankRenderer<MatterCapacitorBankBlockEntity>(context) {
	override fun gaugeLevel(entity: MatterCapacitorBankBlockEntity): Float {
		return entity.gaugeLevel
	}

	override val texture: AbstractMatterySprite
		get() = MatterGaugePanel.GAUGE_FOREGROUND
}

